/*-
 * Copyright (C)2003 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2003 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: open.c,v 1.2 2004/02/02 03:32:57 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "mife_int.h"
#include "mife.h"

static const ssize_t
mindelta(u_int32_t flag) {
    ssize_t s;
    s = 0;
    if  (flag & MIFE_BUFL) s = 1 << ((flag & MIFE_BUFL) >> 4);
    if  (s < MINDELTA) s = MINDELTA;
    return(s);
}

struct mife_descriptor *
mife_open(u_int32_t flag, const char *path) {
    struct mife_descriptor *mife;

    ifBLIN_QV3(flag) fprintf(stderr,"+mife_open flag=%08X, path=%s\n", flag, path);
    if  (!(mife = malloc(sizeof(struct mife_descriptor)))) {
        errno = ENOMEM;
        ifBLIN_QV1(flag) warn("mife_open #1");
        return(NULL);
    }
    bzero(mife, sizeof(struct mife_descriptor));
    mife->flag = flag & ~(MIFE_INMM | MIFE_ALLC | MIFE_MMAP);
    mife->flag |= MIFE_CLOS | MIFE_PIPE;
    mife->mindelta = mindelta(mife->flag);
    if  ((mife->fd = open(path, O_RDONLY)) < 0) {
        ifBLIN_QV1(flag) warn("mife_open #2");
        return(NULL);
    }
    if  ((mife->flag & MIFE_RTRY) == 0) mife->flag |= NRETRY00 & MIFE_RTRY;
    ifBLIN_QV3(flag) fprintf( stderr,"-mife_open flag=%08X, mindelta=%08X, fd=%d\n"
                            , mife->flag, mife->mindelta, mife->fd
                            );
    return(mife);
}

struct mife_descriptor *
mife_opef(u_int32_t flag, int fd) {
    struct mife_descriptor *mife;

    ifBLIN_QV3(flag) fprintf(stderr,"+mife_opef flag=%08X, fd=%d\n", flag, fd);
    if  (!(mife = malloc(sizeof(struct mife_descriptor)))) {
        errno = ENOMEM;
        ifBLIN_QV1(flag) warn("mife_opef #1");
        return(NULL);
    }
    bzero(mife, sizeof(struct mife_descriptor));
    mife->flag = flag & ~(MIFE_INMM | MIFE_ALLC | MIFE_MMAP | MIFE_CLOS);
    mife->flag |= MIFE_PIPE;
    mife->mindelta = mindelta(mife->flag);
    mife->fd = fd;
    if  ((mife->flag & MIFE_RTRY) == 0) mife->flag |= NRETRY00 & MIFE_RTRY;
    ifBLIN_QV3(flag) fprintf( stderr,"-mife_opef flag=%08X, mindelta=%08X, fd=%d\n"
                            , mife->flag, mife->mindelta, mife->fd
                            );
    return(mife);
}

struct mife_descriptor *
mife_opes(u_int32_t flag, const char *src, size_t size) {
    struct mife_descriptor *mife;
    size_t len;
    int i;

    ifBLIN_QV3(flag) fprintf(stderr,"+mife_opes flag=%08X, size=%u\n", flag, size);
    if  (!(mife = malloc(sizeof(struct mife_descriptor)))) {
        errno = ENOMEM;
        ifBLIN_QV1(flag) warn("mife_opes #1");
        return(NULL);
    }
    bzero(mife, sizeof(struct mife_descriptor));
    mife->flag = flag & ~(MIFE_MMAP | MIFE_CLOS);
    mife->flag |= MIFE_INMM | MIFE_ALLC | MIFE_PIPE;
    mife->mindelta = 0;
    mife->fd = -1;
    mife->offset = 0;
    len = size ? size : strlen(src);
    i = (flag & MIFE_FULL) ? 1 : 0;
    mife->actlen = len;
    if  (!(mife->buffer = malloc(mife->actlen + i))) {
        errno = ENOMEM;
        ifBLIN_QV1(flag) warn("mife_opes #2");
        return(NULL);
    }
    bcopy(src, mife->buffer, mife->actlen);
    if  (i) ((char *)(mife->buffer))[mife->actlen] = 0;
    mife->buflen = mife->actlen + i;
    ifBLIN_QV3(flag) fprintf( stderr,"-mife_opes flag=%08X, mindelta=%08X, buflen=%08X, actlen=%08X\n"
                            , mife->flag, mife->mindelta, mife->buflen, mife->actlen
                            );
    return(mife);
}

struct mife_descriptor *
mife_opem(u_int32_t flag, char *src, size_t size) {
    struct mife_descriptor *mife;
    size_t len;
    int i;

    ifBLIN_QV3(flag) fprintf(stderr,"+mife_opem flag=%08X, size=%u\n", flag, size);
    if  (!(mife = malloc(sizeof(struct mife_descriptor)))) {
        errno = ENOMEM;
        ifBLIN_QV1(flag) warn("mife_opem #1");
        return(NULL);
    }
    bzero(mife, sizeof(struct mife_descriptor));
    mife->flag = flag & ~(MIFE_ALLC | MIFE_MMAP | MIFE_CLOS);
    mife->flag |= MIFE_INMM | MIFE_PIPE;
    mife->mindelta = 0;
    mife->fd = -1;
    mife->offset = 0;
    len = size ? size : strlen(src);
    i = (mife->flag & MIFE_FULL) ? 1 : 0;
    mife->buffer = src;
    mife->actlen = len;
    if  (i) ((char *)(mife->buffer))[mife->actlen] = 0;
    mife->buflen = mife->actlen + i;
    ifBLIN_QV3(flag) fprintf( stderr,"-mife_opem flag=%08X, mindelta=%08X, buflen=%08X, actlen=%08X\n"
                            , mife->flag, mife->mindelta, mife->buflen, mife->actlen
                            );
    return(mife);
}
