/*-
 * Copyright (C)2003 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
|                                                                                                     |
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2003 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: read.c,v 1.2 2004/02/02 03:32:57 babolo Exp $";
#endif /* not lint */

#include <machine/limits.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "mife.h"
#include "mife_int.h"
#include "eol.h"

#define COND_NULL_BUF  0x0001
#define COND_BIG_GAP   0x0002  /*      */
#define COND_NEED_REM  0x0004  /*   ,  remalloc   */
#define COND_NEED_MOV  0x0008

#define C_fr 0x01 /* free buffer     */
#define C_zr 0x02 /* empty buffer    */
#define C_al 0x04 /* allocate buffer */
#define C_mv 0x08 /* shift buffer    */
#define C_ra 0x10 /* realloc buffer  */
#define C_in 0x20 /* read to buffer  */

static const u_char nuldecod[] =
{ 0      /* 00 */
, C_al   /* 01 */
, C_in   /* 02 */
, C_al   /* 03 */
};
static const u_char pipdecod[] =
{ C_in                      /* 00 */
, C_in | C_zr | C_al        /* 01 */
, C_in | C_zr               /* 02 */
, C_in | C_zr | C_al        /* 03 */
, C_in | C_ra               /* 04 */
, C_in | C_zr | C_al        /* 05 */
, C_in | C_fr | C_zr | C_al /* 06 */
, C_in | C_zr | C_al        /* 07 */
, C_in | C_mv               /* 08 */
, C_in | C_zr | C_al        /* 09 */
, C_in | C_zr               /* 0A */
, C_in | C_zr | C_al        /* 0B */
, C_in | C_mv | C_ra        /* 0C */
, C_in | C_zr | C_al        /* 0D */
, C_in | C_fr | C_zr | C_al /* 0E */
, C_in | C_zr | C_al        /* 0F */
};

ssize_t
mife_read(struct mife_descriptor *mife, ssize_t needlen, off_t offset, const char *eoln) {
    struct stat sb;
    struct rlimit rlp;
    int retrylim, plb;
    u_int32_t cond;
    u_int32_t control;

    ifBLIN_QV3(mife->flag) fprintf( stderr, "+mife_read %s needlen=%08X offset=%08X\n"
                                  , mife ? "OK" : "NULL"
                                  , needlen
                                  , (size_t)offset
                                  );
    if  (!mife) {
        errno = EINVAL;
        ifBLIN_QV1(mife->flag) warn("mife_read mife=NULL");
        return(-1);
    }
    ifBLIN_QV4(mife->flag) fprintf( stderr
                                  , "mife->flag=%08X fd=%d mindelta=%08X\n"
                                    "mife->actlen=%08X  buflen=%08X offset=%08X\n"
                                  , mife->flag, mife->fd, mife->mindelta
                                  , mife->actlen, mife->buflen, (size_t)mife->offset
                                  );
    if  (needlen == 0) {
        off_t tsz;
        if  (getrlimit(RLIMIT_DATA, &rlp)) {
            ifBLIN_QV1(mife->flag) warn("mife_read #02");
            return(-1);
        }
        tsz = (mife->flag & MIFE_BUFX)
            ? mife->mindelta
            : mife->mindelta + ((rlp.rlim_cur == RLIM_INFINITY) ? DEFMEM : rlp.rlim_cur >> 8);
        needlen = (tsz > SSIZE_MAX) ? SSIZE_MAX : tsz;
    }
    if  (needlen >= SSIZE_MAX) {
        errno = EDOM;
        ifBLIN_QV1(mife->flag) warn("mife_read #03");
        return(-1);
    }
    if  (needlen < 0) {
        errno = EINVAL;
        ifBLIN_QV1(mife->flag) warn("mife_read #04");
        return(-1);
    }
    plb = (mife->flag & MIFE_FULL) ? 1 : 0;
    ifBLIN_QV4(mife->flag) fprintf(stderr, "needlen=%08X plb=%d\n", needlen, plb);
    if  (mife->flag & MIFE_INMM) {
        off_t delta;
        ifBLIN_QV4(mife->flag) fprintf(stderr, " -> null\n");
        delta = offset - mife->offset;
        if  (delta < (off_t)0) {
            errno = ESPIPE;
            ifBLIN_QV1(mife->flag) warn("mife_read #10 null");
            return(-1);
        }
        cond = 0;
        if  (!mife->buffer) cond |= COND_NULL_BUF;
        if  (mife->actlen < delta) cond |= COND_BIG_GAP;
        control = nuldecod[cond];
        if  (control & C_in) {
            errno = ESPIPE;
            ifBLIN_QV1(mife->flag) warn("mife_read #11");
            return(-1);
        }
        if  (control & C_al) {
            errno = EINVAL;
            ifBLIN_QV1(mife->flag) warnx("mife_read #12 pipe: internal software error");
            return(-1);
        }
        ifBLIN_QV3(mife->flag)
            fprintf(stderr, "-mife_read #12 null %08X\n", (unsigned)(mife->actlen - delta));
        return(mife->actlen - delta);
    }
    if  (fstat(mife->fd, &sb)) {
        ifBLIN_QV1(mife->flag) warn("mife_read #19");
        return(-1);
    }
    ifBLIN_QV4(mife->flag) fprintf( stderr, "st_mode=%011o", sb.st_mode);
    switch (sb.st_mode & S_IFMT) {
    case S_IFREG: {
        ifBLIN_QV4(mife->flag) fprintf(stderr, " -> file\n");
    ;   if  (mife->flag & MIFE_INMM) {
            errno = EINVAL;
            ifBLIN_QV1(mife->flag) warnx("mife_read #20 pipe: internal software error");
            return(-1);
        }
        if  (!(mife->flag & (MIFE_ALLC | MIFE_PIPE))) {
            /* XXXX Not yet */
            ssize_t tmplen;
            off_t tmpsz;

            /* remmap */
            if  (mife->buffer && munmap(mife->buffer, mife->buflen) < 0) {
                ifBLIN_QV3(mife->flag) warn("-mife_read #21: file");
                return(-1);
            }
            /* Biggest tail for this file and offset */
            tmpsz = sb.st_size - offset + plb;
            if  (tmpsz <= 0) {
                tmplen = (tmpsz < -SSIZE_MAX) ? -SSIZE_MAX : tmpsz;
                mife->buffer = NULL;
                mife->actlen = 0;
                mife->buflen = 0;
                mife->offset = offset;
                ifBLIN_QV3(mife->flag) fprintf(stderr, "-mife_read #22 file %d\n", tmplen);
                return(tmplen);
            }
            tmplen = ((rlp.rlim_cur == RLIM_INFINITY) ? DEFMEM : rlp.rlim_cur >> 8) + mife->mindelta;
            if  (mife->flag & MIFE_BUFX) tmplen = mife->mindelta;
            if  (tmpsz < tmplen) tmplen = tmpsz;
            mife->buffer = mmap(NULL, tmplen, PROT_READ | PROT_WRITE, MAP_PRIVATE, mife->fd, offset);
            if  (mife->buffer == MAP_FAILED) {
                ifBLIN_QV3(mife->flag) fprintf(stderr, "-mife_read #23: file: %s\n", strerror(errno));
                return(-1);
            }
            mife->actlen = needlen;
            mife->buflen = tmplen;
            mife->offset = offset;
            if  (mife->flag & MIFE_FULL) ((u_char *)(mife->buffer))[mife->actlen] = 0;
            if  (mife->flag & MIFE_FULL && mife->flag & BLIN_VER4)
                fprintf( stderr, "===================\n%s~~~~~~~~~~~~~~~~~~~\n", (u_char *)(mife->buffer));
            ifBLIN_QV3(mife->flag) fprintf(stderr, "-mife_read #24 file %08X\n", tmplen);
            return(tmplen);
    }   }
    case S_IFIFO:
    case S_IFCHR:
    case S_IFBLK:
    case S_IFSOCK: {
        ssize_t nearlen;
        off_t delta;
        struct eoldetector *eo;

    ;   eo = eolinit(mife->flag, eoln, (const char * const * const)&(mife->buffer));
    ;   if  (eoln && !eo) {
    ;       errno = ENOMEM;
    ;       ifBLIN_QV1(mife->flag) warn("mife_read #30 pipe");
    ;       return(-1);
    ;   }
    ;   nearlen = ((needlen > mife->mindelta) ? needlen : mife->mindelta) + plb;
    ;   ifBLIN_QV4(mife->flag) fprintf(stderr, " -> pipe, nearlen=%08X, eomax=%d\n", nearlen, eo ? eo->eomax : -1);
    ;   if  (mife->flag & (MIFE_INMM | MIFE_MMAP)) {
            errno = EINVAL;
            ifBLIN_QV1(mife->flag) warnx("mife_read #31 pipe: internal software error");
            if  (eo) free(eo);
            return(-1);
        }
    ;   do {
        ;   delta = offset - mife->offset;
        ;   if  (delta < (off_t)0) {
        ;       errno = ESPIPE;
        ;       ifBLIN_QV1(mife->flag) warn("mife_read #32 pipe");
        ;       if  (eo) free(eo);
        ;       return(-1);
        ;   }
        ;   cond = 0;
        ;   control = 0;
        ;   if  (!mife->buffer) cond |= COND_NULL_BUF;
        ;   if  (mife->actlen < delta + needlen) {
        ;       if  (mife->buflen < mife->actlen + nearlen) cond |= COND_NEED_REM;
        ;       if  (mife->buflen < delta + MAXPACKT + plb) cond |= COND_NEED_MOV;
        ;       if  (mife->buflen < mife->actlen + MAXPACKT + plb) cond |= COND_NEED_MOV;
        ;       if  (mife->actlen < delta) cond |= COND_BIG_GAP;
        ;       control = pipdecod[cond];
        ;   }
        ;   ifBLIN_QV4(mife->flag)
                fprintf( stderr
                       , "mife->actlen=%08X  buflen=%08X offset=%08X\n"
                         "delta=%08X cond=%02X control=%02X\n"
                       , mife->actlen, mife->buflen, (size_t)mife->offset
                       , (size_t)delta, cond, control
                       );
        ;   if  (control & C_fr) {
        ;       ifBLIN_QV5(mife->flag) fprintf(stderr, "free\n");
        ;       free(mife->buffer);
        ;       eolstop(eo);
        ;       mife->buflen = 0;
        ;       mife->flag &= ~MIFE_ALLC;
        ;   }
        ;   if  (control & C_zr) {
        ;       ifBLIN_QV5(mife->flag) fprintf(stderr, "zero\n");
        ;       eolstop(eo);
        ;       mife->offset += mife->actlen;
        ;       mife->actlen = 0;
        ;   }
        ;   if  (control & C_al) {
        ;       ifBLIN_QV5(mife->flag) fprintf(stderr, "malloc(%08X)\n", (unsigned)nearlen + MAXPACKT);
        ;       eolstop(eo);
        ;       mife->actlen = 0;
        ;       mife->buflen = nearlen + MAXPACKT;
        ;       mife->flag |= MIFE_ALLC;
        ;       if  (!(mife->buffer = malloc(nearlen + MAXPACKT))) {
        ;           errno = ENOMEM;
        ;           ifBLIN_QV1(mife->flag) warn("mife_read #33 pipe");
        ;           if  (eo) free(eo);
        ;           return(-1);
        ;   }   }
        ;   if  (control & C_mv) {
        ;       ifBLIN_QV5(mife->flag) fprintf(stderr, "need move\n");
        ;       bcopy(mife->buffer + delta, mife->buffer, mife->actlen - delta);
        ;       eoldelta(eo, -delta);
        ;       mife->offset += delta;
        ;       mife->actlen -= delta;
        ;       delta = 0;
        ;   }
        ;   if  (control & C_ra) {
        ;       ifBLIN_QV5(mife->flag)
                    fprintf(stderr, "reallocf(%08X)\n", nearlen + mife->actlen + MAXPACKT);
        ;       mife->buflen = nearlen + mife->actlen + MAXPACKT;
        ;       if  (!(mife->buffer = reallocf(mife->buffer, nearlen + mife->actlen + MAXPACKT))) {
        ;           errno = EFBIG;
        ;           ifBLIN_QV1(mife->flag) warn("mife_read #34 pipe");
        ;           if  (eo) free(eo);
        ;           return(-1);
        ;   }   }
        ;   if  (eoldetect(mife->flag, eo, mife->actlen) >= 0) {
        ;       ifBLIN_QV4(mife->flag) fprintf(stderr, "p eolp=%d, eolst=%d\n", eo ? eo->eolp : -1, eo ? eo->eolst : -1);
        ;       control = 0;
        ;   }
        ;   if  (control & C_in) {
                ssize_t tmplen;
        ;       ifBLIN_QV5(mife->flag)
                    fprintf( stderr, "input: delta=%08X buflen=%08X needlen=%08X\n"
                           , (size_t)delta, mife->buflen, needlen
                           );
        ;       retrylim = mife->flag & MIFE_RTRY;
        ;       do {
        ;           tmplen = read( mife->fd
                                 , mife->buffer + mife->actlen
                                 , (needlen > MAXPACKT) ? needlen : MAXPACKT
                                 );
        ;           ifBLIN_QV6(mife->flag)
                        fprintf( stderr, "retrylim=%d tmplen=%08X nearlen=%08X\n"
                               , retrylim, tmplen, nearlen
                               );
        ;           if  (tmplen < 0) {
        ;               if  (errno == EAGAIN || errno == EDEADLK) {
        ;                   ifBLIN_QV2(mife->flag) warn("mife_read #35 pipe");
        ;               } else {
        ;                   ifBLIN_QV1(mife->flag) warn("mife_read #36 pipe");
        ;                   if  (eo) free(eo);
        ;                   return(-1);
        ;               }
        ;           } else if (tmplen == 0) {
        ;               if  (retrylim-- <= 0 || S_ISREG(sb.st_mode)) break;
        ;               usleep(1 << ((mife->flag & MIFE_RTRY) - retrylim));
        ;           } else {
        ;               mife->actlen += tmplen;
        ;               eolstart(eo, delta);
        ;               if  (eoldetect(mife->flag, eo, mife->actlen) >= 0) {
        ;                   ifBLIN_QV5(mife->flag)
                                fprintf(stderr, " eolp=%d, eolst=%d\n", eo ? eo->eolp : -1, eo ? eo->eolst : -1);
        ;                   control = 0;
        ;                   break;
        ;           }   }
        ;       } while (!tmplen);
        ;       if  (!tmplen) {
        ;           mife->flag |= MIFE_EOFL;
        ;           break;
        ;   }   }
        } while (control);
    ;   {   ssize_t tmplen;
        ;   tmplen = mife->buflen - mife->actlen;
        ;   delta = offset - mife->offset;
        ;   ifBLIN_QV4(mife->flag)
                fprintf( stderr
                       , "mife->actlen=%08X  buflen=%08X offset=%08X delta=%08X\n"
                       , mife->actlen, mife->buflen, (size_t)mife->offset, (size_t)delta
                       );
        ;   if  (tmplen <= 0) {
        ;       errno = EINVAL;
        ;       ifBLIN_QV1(mife->flag) warnx("mife_read #37 pipe: internal software error");
        ;       if  (eo) free(eo);
        ;       return(-1);
        ;   }
        ;   if  (delta > mife->actlen) {
        ;       errno = ESPIPE;
        ;       ifBLIN_QV1(mife->flag) warn("mife_read #38 pipe");
        ;       if  (eo) free(eo);
        ;       return(-1);
        ;   }
        ;   if  (tmplen > 0) bzero(mife->buffer + mife->actlen, (tmplen > 8) ? 8 : tmplen);
        ;   ifBLIN_QV3(mife->flag) fprintf( stderr, "-mife_read #39 pipe %08X(%d)\n"
                                          , (unsigned)(mife->actlen - delta), (int)tmplen
                                          );
        ;   if  (eo) free(eo);
        ;   return((ssize_t)mife->actlen - delta);
    }   }
    case S_IFDIR:
    case S_IFLNK:
    case S_IFWHT:
    default:
        errno = EINVAL;
        ifBLIN_QV1(mife->flag) warn(" -> unknown\n-mife_read #40");
        return(-1);
}   }
